"""
The MIT License (MIT)

Copyright (c) [2015-2022] [Andrew Annex]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Sources for this file are mostly from DaRasch, spiceminer/getcspice.py,
with edits by me as needed for python2/3 compatibility
https://github.com/DaRasch/spiceminer/blob/master/getcspice.py

The MIT License (MIT)

Copyright (c) 2013 Philipp Rasch

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

The MIT License (MIT)

Copyright (c) 2017 ODC Space

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
import io
import platform
import os
from pathlib import Path
import shutil
import subprocess
import sys
import time
import tempfile
import urllib
import urllib.request
import urllib.error
from zipfile import ZipFile

CSPICE_SRC_DIR = "CSPICE_SRC_DIR"
CSPICE_SHARED_LIB = "CSPICE_SHARED_LIB"
CSPICE_NO_PATCH = "CSPICE_NO_PATCH"

host_OS = platform.system()
# Check if platform is supported
os_supported = host_OS in ("Linux", "Darwin", "FreeBSD", "Windows")
# Get platform is Unix-like OS or not
is_unix = host_OS in ("Linux", "Darwin", "FreeBSD")
# Get current working directory
root_dir = str(Path(os.path.realpath(__file__)).parent)
# Make the directory path for cspice
cspice_dir = os.environ.get(CSPICE_SRC_DIR, os.path.join(root_dir, "cspice"))
# and make a global tmp cspice directory
tmp_cspice_root_dir = None
# versions
spice_version = "N0067"
spice_num_v = "67"


class GetCSPICE(object):
    """
    Class that support the download from the NAIF FTP server of the required
    CSPICE package for the architecture used by the Python distribution that
    invokes this module.  By default the CSPICE Toolkit version N0066 is
    downloaded and unpacked on the directory where this module is located.

    Arguments
    ---------
    :argument version: String indicating the required version of the CSPICE
                       Toolkit. By default it is 'N0066'.
    :type: str

    """

    # This class variable will be used to store the CSPICE package in memory.
    _local = None

    # Supported distributions
    _dists = {
        # system   arch        distribution name           extension
        # -------- ----------  -------------------------   ---------
        ("Darwin", "32bit"): ("MacIntel_OSX_AppleC_32bit", "tar.Z"),
        ("Darwin", "64bit"): ("MacIntel_OSX_AppleC_64bit", "tar.Z"),
        ("Darwin", "64bit"): ("MacM1_OSX_clang_64bit", "tar.Z"),
        ("cygwin", "32bit"): ("PC_Cygwin_GCC_32bit", "tar.Z"),
        ("cygwin", "64bit"): ("PC_Cygwin_GCC_64bit", "tar.Z"),
        ("FreeBSD", "32bit"): ("PC_Linux_GCC_32bit", "tar.Z"),
        ("FreeBSD", "64bit"): ("PC_Linux_GCC_64bit", "tar.Z"),
        ("Linux", "32bit"): ("PC_Linux_GCC_32bit", "tar.Z"),
        ("Linux", "64bit"): ("PC_Linux_GCC_64bit", "tar.Z"),
        ("Windows", "32bit"): ("PC_Windows_VisualC_32bit", "zip"),
        ("Windows", "64bit"): ("PC_Windows_VisualC_64bit", "zip"),
    }

    def __init__(self, version=spice_version, dst=None):
        """Init method that uses either the default N0066 toolkit version token
        or a user provided one.
        """
        try:
            # Get the remote file path for the Python architecture that
            # executes the script.
            distribution, self._ext = self._distribution_info()
        except KeyError:
            print("SpiceyPy currently does not support your system.")
        else:
            cspice = "cspice.{}".format(self._ext)
            self._rcspice = (
                "https://naif.jpl.nasa.gov/pub/naif/misc"
                "/toolkit_{0}/C/{1}/packages"
                "/{2}"
            ).format(version, distribution, cspice)

            # Setup the local directory (where the package will be downloaded)
            if dst is None:
                dst = os.path.realpath(os.path.dirname(__file__))
            self._root = dst

            # Download the file
            print("Downloading CSPICE for {0}...".format(distribution))
            attempts = 10  # Let's try a maximum of attempts for getting SPICE
            while attempts:
                attempts -= 1
                try:
                    self._download()
                except RuntimeError as error:
                    print(
                        "Download failed with URLError: {0}, trying again after "
                        "15 seconds!".format(error)
                    )
                    time.sleep(15)
                else:
                    # Unpack the file
                    print("Unpacking... (this may take some time!)")
                    self._unpack()
                    # We are done.  Let's return to the calling code.
                    break

    def _distribution_info(self):
        """Creates the distribution name and the expected extension for the
        CSPICE package and returns it.

        :return (distribution, extension) tuple where distribution is the best
                guess from the strings available within the platform_urls list
                of strings, and extension is either "zip" or "tar.Z" depending
                on whether we are dealing with a Windows platform or else.
        :rtype: tuple (str, str)

        :raises: KeyError if the (system, machine) tuple does not correspond
                 to any of the supported SpiceyPy environments.
        """

        print("Gathering information...")
        system = platform.system()

        # Cygwin system is CYGWIN-NT-xxx.
        system = "cygwin" if "CYGWIN" in system else system

        processor = platform.processor()
        machine = "64bit" if sys.maxsize > 2 ** 32 else "32bit"
        machine2 = platform.machine()

        print("SYSTEM:   ", system)
        print("PROCESSOR:", processor)
        print("MACHINE:  ", machine, machine2)

        return self._dists[(system, machine)]

    def _download(self):
        """Support function that encapsulates the OpenSSL transfer of the CSPICE
        package to the self._local io.ByteIO stream.

        :raises RuntimeError if there has been any issue with the HTTPS
                             communication

        .. note::

           Handling of CSPICE downloads from HTTPS
           ---------------------------------------
           Some Python distributions may be linked to an old version of OpenSSL
           which will not let you connect to NAIF server due to recent SSL cert
           upgrades on the JPL servers.  Moreover, versions older than
           OpenSSL 1.0.1g are known to contain the 'the Heartbleed Bug'.
           Therefore this method provides two different implementations for the
           HTTPS GET call to the NAIF server to download the required CSPICE
           distribution package.

           * as of 3.0.1, we default back to use built in openssl, as we require python 3.6 or above *
        """
        try:
            # Send the request to get the CSPICE package (proxy auto detected).
            response = urllib.request.urlopen(self._rcspice, timeout=10)
        except urllib.error.URLError as err:
            raise RuntimeError(err.reason)

        # Convert the response to io.BytesIO and store it in local memory.
        self._local = io.BytesIO(response.read())

    def _unpack(self):
        """Unpacks the CSPICE package on the given root directory. Note that
        Package could either be the zipfile.ZipFile class for Windows platforms
        or tarfile.TarFile for other platforms.
        """
        if self._ext == "zip":
            with ZipFile(self._local, "r") as archive:
                archive.extractall(self._root)
        else:
            cmd = (
                f"gunzip | tar {'xfC -' if host_OS == 'FreeBSD' else 'xC'} {self._root}"
            )
            proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE)
            proc.stdin.write(self._local.read())
        self._local.close()


def copy_supplements() -> None:
    """
    Copy supplement files (patches, windows build files)
    to cspice directory
    """
    cwd = os.getcwd()
    patches = list(Path().cwd().glob("*.patch"))
    print("copy supplements to: ", cspice_dir, flush=True)
    for p in patches:
        shutil.copy(p, cspice_dir)
    if host_OS == "Windows":
        windows_files = [Path("./makeDynamicSpice.bat"), Path("./cspice.def")]
        cspice_src_dir = os.path.join(cspice_dir, "cspice", "src", "cspice")
        for w in windows_files:
            shutil.copy(w, cspice_src_dir)
    os.chdir(cwd)
    pass


def apply_patches() -> None:
    if int(spice_num_v) == 66:
        cwd = os.getcwd()
        os.chdir(cspice_dir)
        iswin = "-windows" if host_OS == "Windows" else ""
        patches = [
            f"0001-patch-for-n66-dskx02.c{iswin}.patch",
            f"0002-patch-for-n66-subpnt.c{iswin}.patch",
        ]
        if (
            host_OS == "Darwin"
        ):  # todo is this only needed for M1 or is it for any macos
            patches.append("0004_inquire_unistd.patch")
        for p in patches:
            try:
                print(f"Applying Patch {p}", flush=True)
                patch_cmd = subprocess.run(["git", "apply", "--reject", p], check=True)
            except subprocess.CalledProcessError as cpe:
                raise cpe
        os.chdir(cwd)
    pass


def prepare_cspice() -> None:
    """
    Prepare temporary cspice source directory,
    If not provided by the user, or if not readable, download a fresh copy
    :return: None
    """
    cwd = os.getcwd()
    global cspice_dir, tmp_cspice_root_dir
    with tempfile.TemporaryDirectory(prefix="cspice_spiceypy_") as tmp_dir:
        # Trick for python <3.8, delete tmp dir so that we can write it overwrite it
        tmp_cspice_root_dir = str((Path(tmp_dir)).absolute())
    tmp_cspice_src_dir = os.path.join(tmp_cspice_root_dir, "cspice")
    if os.access(cspice_dir, os.R_OK):
        print("Found usable CSPICE src directory", flush=True)
        # newer shutil.copytree has a flag for this that negates need for this
        shutil.copytree(
            cspice_dir + "/",
            tmp_cspice_src_dir,
            copy_function=shutil.copy,
        )
    else:
        print("Downloading CSPICE src from NAIF", flush=True)
        Path(tmp_cspice_src_dir).mkdir(exist_ok=True, parents=True)
        GetCSPICE(dst=tmp_cspice_root_dir)
    # okay cspice_dir need to be inside the root level cspice src dir
    cspice_dir = tmp_cspice_root_dir
    print("end of prep:", cspice_dir, flush=True)
    os.chdir(cwd)
    # okay now copy any and all files needed for building
    pass


def build_cspice() -> str:
    """
    Builds cspice
    :return: absolute path to new compiled shared library
    """
    cwd = os.getcwd()
    global cspice_dir, host_OS
    if is_unix:
        libname = f"libcspice.so"
        if host_OS == "Darwin":
            extra_flags = f"-dynamiclib -install_name @rpath/{libname}"
        else:
            extra_flags = f"-shared -Wl,-soname,{libname}"
        destination = cspice_dir
        os.chdir(destination)
        cmds = [
            "gcc -Iinclude -c -fPIC -O2 -ansi ./cspice/src/cspice/*.c",
            f"gcc {extra_flags} -fPIC -O2 -lm *.o -o {libname}",
        ]
    elif host_OS == "Windows":
        destination = os.path.join(cspice_dir, "cspice", "src", "cspice")
        os.chdir(destination)
        cmds = ["makeDynamicSpice.bat"]
    else:
        os.chdir(cwd)
        raise NotImplementedError(f"non implemented host os for build {host_OS}")
    try:
        for cmd in cmds:
            _ = subprocess.run(cmd, shell=True, check=True)
    except subprocess.CalledProcessError as cpe:
        os.chdir(cwd)
        pass
    # get the built shared library
    shared_lib_path = [
        str(p.absolute())
        for p in Path(destination).glob("*.*")
        if p.suffix in (".dll", ".66", ".dylib", ".so")
    ]
    if len(shared_lib_path) != 1:
        os.chdir(cwd)
        raise RuntimeError(
            f'Could not find built shared library of SpiceyPy in {list(Path(destination).glob("*.*"))}'
        )
    else:
        shared_lib_path = shared_lib_path[0]
    print(shared_lib_path, flush=True)
    os.chdir(cwd)
    return shared_lib_path


def main() -> None:
    """
    Main routine to build or not build cspice
    expected tmp src dir layout
    /tmpdir/ < top level temporary directory from TemporaryDirectory
           /cspice/ < directory containing contents of source code for cspice
                  /lib
                  /bin
                  /src/cspice/
                  ...
    :return: None
    """
    cwd = os.getcwd()
    # set final destination for cspice dynamic library
    destination = os.path.join(
        root_dir,
        "src",
        "spiceypy",
        "utils",
        "libcspice.so" if is_unix else "libcspice.dll",
    )
    # check if the shared library already exists, if it does we are done
    if Path(destination).is_file():
        print(
            "Done! shared library for cspice already exists in destination. Done!",
            flush=True,
        )
        return
    # next see if cspice shared library is provided
    shared_library_path = os.environ.get(CSPICE_SHARED_LIB)
    if shared_library_path is not None:
        print(f"User has provided a shared library...", flush=True)
        # todo: what if we can't read the file? we need to jump to building it... doubtful this happens
        pass  # now we don't need to do that much
    else:
        # okay, now we either are given a src dir, have already downloaded it, or don't have it
        print("Preparing cspice", flush=True)
        prepare_cspice()
        # add the patches
        print("Copying supplements", flush=True)
        copy_supplements()
        # okay now that we have the source in a writeable directory we can apply patches
        if CSPICE_NO_PATCH not in os.environ:
            print("Apply patches", flush=True)
            apply_patches()
        # now build
        print("Building cspice", flush=True)
        shared_library_path = build_cspice()
    print(f"Copying built cspice: {shared_library_path} to {destination}", flush=True)
    # first make the directory for the destination if it doesn't exist
    Path(destination).parent.mkdir(parents=True, exist_ok=True)
    # okay now move shared library to dst dir
    shutil.copyfile(shared_library_path, destination)
    # cleanup tmp dir, windows seems to fail with this:
    #    PermissionError: [WinError 32] The process cannot access the file because it is being used by another process
    # if tmp_cspice_root_dir is not None:
    #     if os.path.exists(tmp_cspice_root_dir) and os.path.isdir(tmp_cspice_root_dir):
    #         shutil.rmtree(tmp_cspice_root_dir)
    # and now we are done!
    os.chdir(cwd)
    print("Done!")


if __name__ == "__main__":
    main()



"""
The MIT License (MIT)

Copyright (c) [2015-2022] [Andrew Annex]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

The MIT License (MIT)

Copyright (c) 2013 Philipp Rasch

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import collections.abc as collections_abc

from array import array
from typing import Type
from ctypes import (
    c_char_p,
    c_int,
    c_double,
    c_char,
    c_void_p,
    sizeof,
    Array,
    create_string_buffer,
    cast,
    Structure,
    string_at,
    POINTER,
)

import numpy
from numpy import ctypeslib as numpc

from .exceptions import (
    SpiceyError,
    SpiceyPyError,
    NotFoundError,
    SpiceyPyIOError,
    SpiceyPyMemoryError,
    SpiceyPyTypeError,
    SpiceyPyKeyError,
    SpiceyPyIndexError,
    SpiceyPyRuntimeError,
    SpiceyPyZeroDivisionError,
    SpiceyPyValueError,
    exceptions,
)


# Collection of supporting functions for wrapper functions
__author__ = "AndrewAnnex"


def short_to_spiceypy_exception_class(short: str) -> Type[SpiceyError]:
    """
    Lookup the correct Spice Exception class

    :param short: Spice error system short description key
    :return: SpiceyError
    """
    return exceptions.get(short, SpiceyError)


def dynamically_instantiate_spiceyerror(
    short: str = "",
    explain: str = "",
    long: str = "",
    traceback: str = "",
    found: str = "",
):
    """
    Dynamically creates a SpiceyPyException which is a subclass of SpiceyError and
    may also be subclassed to other exceptions such as IOError and such depending on the Short description

    :param short:
    :param explain:
    :param long:
    :param traceback:
    :param found:
    :return:
    """
    base_exception = short_to_spiceypy_exception_class(short)
    return base_exception(
        short=short, explain=explain, long=long, traceback=traceback, found=found
    )


def to_double_vector(x):
    return DoubleArray.from_param(param=x)


def to_double_matrix(x):
    return DoubleMatrix.from_param(param=x)


def to_int_vector(x):
    return IntArray.from_param(param=x)


def to_int_matrix(x):
    return IntMatrix.from_param(param=x)


def is_iterable(i) -> bool:
    """
    From stackoverflow
    https://stackoverflow.com/questions/1055360/how-to-tell-a-variable-is-iterable-but-not-a-string/44328500#44328500
    :param i: input collection
    :return: if the input is iterable but not a string
    """
    return isinstance(i, collections_abc.Iterable) and not isinstance(i, str)


def to_python_string(in_string):
    if isinstance(in_string, c_char_p):
        return to_python_string(in_string.value)
    else:
        return bytes.decode(string_at(in_string), errors="ignore").rstrip()


def empty_char_array(x_len=None, y_len=None):
    if not y_len:
        y_len = 1
    if not x_len:
        x_len = 1
    if isinstance(x_len, c_int):
        x_len = x_len.value
    if isinstance(y_len, c_int):
        y_len = y_len.value
    return ((c_char * x_len) * y_len)()


def empty_double_matrix(x=3, y=3):
    if isinstance(x, c_int):
        x = x.value
    if isinstance(y, c_int):
        y = y.value
    return ((c_double * x) * y)()


def empty_double_vector(n):
    if isinstance(n, c_int):
        n = n.value
    assert isinstance(n, int)
    return (c_double * n)()


def empty_int_matrix(x=3, y=3):
    if isinstance(x, c_int):
        x = x.value
    if isinstance(y, c_int):
        y = y.value
    return ((c_int * x) * y)()


def empty_int_vector(n):
    if isinstance(n, c_int):
        n = n.value
    assert isinstance(n, int)
    return (c_int * n)()


def c_vector_to_python(x):
    """
    Convert the c vector data into the correct python data type
    (numpy arrays or strings)

    :param x: ctypes array
    :return: Iterable
    """
    if isinstance(x[0], bool):
        return numpy.frombuffer(x, dtype=numpy.bool).copy()
    elif isinstance(x[0], int):
        return numpy.frombuffer(x, dtype=numpy.int32).copy()
    elif isinstance(x[0], float):
        return numpy.frombuffer(x, dtype=numpy.float64).copy()
    elif isinstance(x[0].value, bytes):
        return [to_python_string(y) for y in x]


def c_int_vector_to_bool_python(x):
    return numpc.as_array(x).copy().astype(bool)


def c_matrix_to_numpy(x):
    """
    Convert a ctypes 2d array (or matrix) into a numpy array for python use

    :param x: thing to convert
    :return: numpy.ndarray
    """
    return numpc.as_array(x).copy()


def string_to_char_p(inobject, inlen=None):
    """
    convert a python string to a char_p

    :param inobject: input string, int for getting null string of length of int
    :param inlen: optional parameter, length of a given string can be specified
    :return:
    """
    if inlen and isinstance(inobject, str):
        return create_string_buffer(inobject.encode(encoding="UTF-8"), inlen)
    if isinstance(inobject, bytes):
        return inobject
    if isinstance(inobject, c_int):
        return string_to_char_p(" " * inobject.value)
    if isinstance(inobject, int):
        return string_to_char_p(" " * inobject)
    if isinstance(inobject, numpy.str_):
        return c_char_p(inobject.encode(encoding="utf-8"))
    return c_char_p(inobject.encode(encoding="UTF-8"))


def list_to_char_array(arg, x_len=None, y_len=None):
    assert is_iterable(arg)
    if not y_len:
        y_len = len(arg)
    if not x_len:
        x_len = max(len(s) for s in arg) + 1
    if isinstance(x_len, c_int):
        x_len = x_len.value
    if isinstance(y_len, c_int):
        y_len = y_len.value
    return ((c_char * x_len) * y_len)(*[string_to_char_p(l, inlen=x_len) for l in arg])


def list_to_char_array_ptr(input, x_len=None, y_len=None):
    return cast(list_to_char_array(input, x_len=x_len, y_len=y_len), c_char_p)


class DoubleArrayType:
    """
    Class type that will handle all double vectors,
    inspiration from python cookbook 3rd edition
    """

    def from_param(self, param):
        if hasattr(param, "__array_interface__"):
            return self.from_ndarray(param)
        elif isinstance(param, Array):
            return param
        elif isinstance(param, (list, tuple)):
            return self.from_list(param)
        elif isinstance(param, array):
            if param.typecode != "d":
                raise TypeError("must be an array of doubles")
            return self.from_list(param)
        else:
            raise TypeError(f"Can't convert {type(param)}")

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_double) * len(param))(*param)
        return val

    # Cast from a numpy array,
    def from_ndarray(self, param):
        # return param.data_as(POINTER(c_double))
        # the above older method does not work with
        # functions which take vectors of known size
        return numpc.as_ctypes(
            param.astype(numpy.float64, casting="same_kind", copy=False)
        )


class DoubleMatrixType:
    """
    Class type that will handle all 2d double arrays,
    inspiration from python cookbook 3rd edition
    """

    def from_param(self, param):
        if hasattr(param, "__array_interface__"):
            return self.from_ndarray(param)
        elif isinstance(param, Array):
            return param
        elif isinstance(param, (list, tuple)):
            return self.from_list(param)
        else:
            raise TypeError(f"Can't convert {type(param)}")

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_double * len(param[0])) * len(param))(
            *[DoubleArray.from_param(x) for x in param]
        )
        return val

    # Cast from a numpy array
    def from_ndarray(self, param):
        return numpc.as_ctypes(
            param.astype(numpy.float64, casting="same_kind", copy=False)
        )


class IntArrayType:
    """
    Class type that will handle all int vectors,
    inspiration from python cookbook 3rd edition
    """

    def from_param(self, param):
        if hasattr(param, "__array_interface__"):
            return self.from_ndarray(param)
        elif isinstance(param, Array):
            return param
        elif isinstance(param, (list, tuple)):
            return self.from_list(param)
        elif isinstance(param, array):
            if param.typecode != "i":
                raise TypeError("must be an array of ints")
            return self.from_list(param)
        else:
            raise TypeError(f"Can't convert {type(param)}")

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_int) * len(param))(*param)
        return val

    # Cast from a numpy array
    def from_ndarray(self, param):
        # cspice always uses a int size half as big as the float, ie int32 if a float64 system default
        return numpc.as_ctypes(
            param.astype(numpy.int32, casting="same_kind", copy=False)
        )


class IntMatrixType:
    """
    Class type that will handle all 2d int arrays,
    inspiration from python cookbook 3rd edition
    """

    def from_param(self, param):
        if hasattr(param, "__array_interface__"):
            return self.from_ndarray(param)
        elif isinstance(param, Array):
            return param
        elif isinstance(param, (list, tuple)):
            return self.from_list(param)
        else:
            raise TypeError(f"Can't convert {type(param)}")

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_int * len(param[0])) * len(param))(
            *[IntArray.from_param(x) for x in param]
        )
        return val

    # Cast from a numpy array
    def from_ndarray(self, param):
        # cspice always uses a int size half as big as the float, ie int32 if a float64 system default
        return numpc.as_ctypes(
            param.astype(numpy.int32, casting="same_kind", copy=False)
        )


DoubleArray = DoubleArrayType()

DoubleMatrix = DoubleMatrixType()

IntArray = IntArrayType()

IntMatrix = IntMatrixType()


class Plane(Structure):
    _fields_ = [("_normal", c_double * 3), ("_constant", c_double)]

    @property
    def normal(self):
        return c_vector_to_python(self._normal)

    @property
    def constant(self):
        return self._constant

    def __str__(self) -> str:
        return f"<SpicePlane: normal={', '.join([str(x) for x in self._normal])}; constant={self._constant}>"


class Ellipse(Structure):
    _fields_ = [
        ("_center", c_double * 3),
        ("_semi_major", c_double * 3),
        ("_semi_minor", c_double * 3),
    ]

    @property
    def center(self):
        return c_vector_to_python(self._center)

    @property
    def semi_major(self):
        return c_vector_to_python(self._semi_major)

    @property
    def semi_minor(self):
        return c_vector_to_python(self._semi_minor)

    def __str__(self) -> str:
        return f"<SpiceEllipse: center = {self.center} semi_major = {self.semi_major}, semi_minor = {self.semi_minor}>"


class DataType(object):
    SPICE_CHR = 0
    SPICE_DP = 1
    SPICE_INT = 2
    SPICE_TIME = 3
    SPICE_BOOL = 4
    CHR = 0
    DP = 1
    INT = 2
    TIME = 3
    BOOL = 4

    def __init__(self) -> None:
        pass


class SpiceDSKDescr(Structure):
    _fields_ = [
        ("_surfce", c_int),
        ("_center", c_int),
        ("_dclass", c_int),
        ("_dtype", c_int),
        ("_frmcde", c_int),
        ("_corsys", c_int),
        ("_corpar", c_double * 10),
        ("_co1min", c_double),
        ("_co1max", c_double),
        ("_co2min", c_double),
        ("_co2max", c_double),
        ("_co3min", c_double),
        ("_co3max", c_double),
        ("_start", c_double),
        ("_stop", c_double),
    ]

    @property
    def surfce(self):
        return self._surfce

    @property
    def center(self):
        return self._center

    @property
    def dclass(self):
        return self._dclass

    @property
    def dtype(self):
        return self._dtype

    @property
    def frmcde(self):
        return self._frmcde

    @property
    def corsys(self):
        return self._corsys

    @property
    def corpar(self):
        return c_vector_to_python(self._corpar)

    @property
    def co1min(self):
        return self._co1min

    @property
    def co1max(self):
        return self._co1max

    @property
    def co2min(self):
        return self._co2min

    @property
    def co2max(self):
        return self._co2max

    @property
    def co3min(self):
        return self._co3min

    @property
    def co3max(self):
        return self._co3max

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop


class SpiceDLADescr(Structure):
    _fields_ = [
        ("_bwdptr", c_int),
        ("_fwdptr", c_int),
        ("_ibase", c_int),
        ("_isize", c_int),
        ("_dbase", c_int),
        ("_dsize", c_int),
        ("_cbase", c_int),
        ("_csize", c_int),
    ]

    @property
    def bwdptr(self):
        return self._bwdptr

    @property
    def fwdptr(self):
        return self._fwdptr

    @property
    def ibase(self):
        return self._ibase

    @property
    def isize(self):
        return self._isize

    @property
    def dbase(self):
        return self._dbase

    @property
    def dsize(self):
        return self._dsize

    @property
    def cbase(self):
        return self._cbase

    @property
    def csize(self):
        return self._csize


class SpiceEKDataType(c_int):
    _SPICE_CHR = c_int(0)
    _SPICE_DP = c_int(1)
    _SPICE_INT = c_int(2)
    _SPICE_TIME = c_int(3)
    _SPICE_BOOL = c_int(4)

    _fields_ = [
        ("SPICE_CHR", _SPICE_CHR),
        ("SPICE_DP", _SPICE_DP),
        ("SPICE_INT", _SPICE_INT),
        ("SPICE_TIME", _SPICE_TIME),
        ("SPICE_BOOL", _SPICE_BOOL),
    ]

    SPICE_CHR = _SPICE_CHR.value
    SPICE_DP = _SPICE_DP.value
    SPICE_INT = _SPICE_INT.value
    SPICE_TIME = _SPICE_TIME.value
    SPICE_BOOL = _SPICE_BOOL.value


def empty_spice_ek_data_type_vector(n):
    if isinstance(n, c_int):
        n = n.value
    assert isinstance(n, int)
    return (SpiceEKDataType * n)()


class SpiceEKExprClass(c_int):
    _SPICE_EK_EXP_COL = c_int(0)
    _SPICE_EK_EXP_FUNC = c_int(1)
    _SPICE_EK_EXP_EXPR = c_int(2)

    _fields_ = [
        ("SPICE_EK_EXP_COL", _SPICE_EK_EXP_COL),
        ("SPICE_EK_EXP_FUNC", _SPICE_EK_EXP_FUNC),
        ("SPICE_EK_EXP_EXPR", _SPICE_EK_EXP_EXPR),
    ]

    SPICE_EK_EXP_COL = _SPICE_EK_EXP_COL.value
    SPICE_EK_EXP_FUNC = _SPICE_EK_EXP_FUNC.value
    SPICE_EK_EXP_EXPR = _SPICE_EK_EXP_EXPR.value


class SpiceSPK18Subtype(c_int):
    _S18TP0 = c_int(0)
    _S18TP1 = c_int(1)
    S18TP0 = _S18TP0.value
    S18TP1 = _S18TP1.value
    _fields_ = [("S18TP0", _S18TP0), ("S18TP1", _S18TP1)]


def empty_spice_ek_expr_class_vector(n):
    if isinstance(n, c_int):
        n = n.value
    assert isinstance(n, int)
    return (SpiceEKExprClass * n)()


class SpiceEKAttDsc(Structure):
    _fields_ = [
        ("_cclass", c_int),
        ("_dtype", SpiceEKDataType),
        ("_strlen", c_int),
        ("_size", c_int),
        ("_indexd", c_int),
        ("_nullok", c_int),
    ]

    @property
    def cclass(self):
        return self._cclass

    @property
    def dtype(self):
        return self._dtype.value

    @property
    def strlen(self):
        return self._strlen

    @property
    def size(self):
        return self._size

    @property
    def indexd(self):
        return bool(self._indexd)

    @property
    def nullok(self):
        return bool(self._nullok)

    def __str__(self) -> str:
        return (
            f"<SpiceEKAttDsc cclass = {self.cclass}, dtype = {self.dtype}, strlen = {self.strlen}, "
            f"size = {self.size}, indexd = {self.indexd}, nullok = {self.nullok} >"
        )


class SpiceEKSegSum(Structure):
    _fields_ = [
        ("_tabnam", c_char * 65),
        ("_nrows", c_int),
        ("_ncols", c_int),
        ("_cnames", (c_char * 100) * 33),
        ("_cdescrs", SpiceEKAttDsc * 100),
    ]

    @property
    def tabnam(self):
        return to_python_string(self._tabnam)

    @property
    def nrows(self):
        return self._nrows

    @property
    def ncols(self):
        return self._ncols

    @property
    def cnames(self):
        return c_vector_to_python(self._cnames)[0 : self.ncols]

    @property
    def cdescrs(self):
        return self._cdescrs[0 : self.ncols]

    def __str__(self) -> str:
        return (
            f"<SpiceEKSegSum tabnam = {self.tabnam}, nrows = {self.nrows}, ncols = {self.ncols},"
            f" cnames = {self.cnames}, cdescrs = {self.cdescrs} >"
        )


# SpiceCell implementation below is inpart from github.com/DaRasch/spiceminer/
# and modified as needed for this author, maybe we should work together?

# helper classes/functions
BITSIZE = {
    "char": sizeof(c_char),
    "int": sizeof(c_int),
    "double": sizeof(c_double),
    "bool": sizeof(c_int),
    "time": sizeof(c_int),
}


def _char_getter(data_p, index, length):
    return to_python_string(
        (c_char * length).from_address(data_p + index * length * BITSIZE["char"])
    )


def _double_getter(data_p, index, length):
    return c_double.from_address(data_p + index * BITSIZE["double"]).value


def _int_getter(data_p, index, length):
    return c_int.from_address(data_p + index * BITSIZE["int"]).value


class SpiceCell(Structure):
    # Most written by DaRasch, see included MIT license at file header
    DATATYPES_ENUM = {"char": 0, "double": 1, "int": 2, "time": 3, "bool": 4}
    DATATYPES_GET = [_char_getter, _double_getter] + [_int_getter] * 3
    baseSize = 6
    minCharLen = 6
    CTRLBLOCK = 6
    _fields_ = [
        ("dtype", c_int),
        ("length", c_int),
        ("size", c_int),
        ("card", c_int),
        ("isSet", c_int),
        ("adjust", c_int),
        ("init", c_int),
        ("base", c_void_p),
        ("data", c_void_p),
    ]

    def __init__(
        self,
        dtype=None,
        length=None,
        size=None,
        card=None,
        isSet=None,
        base=None,
        data=None,
    ):
        super(SpiceCell, self).__init__()
        self.dtype = dtype
        self.length = length
        self.size = size
        self.card = card
        self.isSet = isSet
        self.adjust = 0  # Always False, because not implemented
        self.init = 0  # Always False, because this is the constructor
        self.base = base  # void pointer
        self.data = data

    def __str__(self) -> str:
        return (
            f"<SpiceCell dtype = {self.dtype}, length = {self.length}, size = {self.size}, card = {self.card},"
            f" is_set = {self.isSet}, adjust = {self.adjust}, init = {self.init}, base = {self.base}, data = {self.data}>"
        )

    def is_int(self):
        return self.dtype == 2

    def is_double(self):
        return self.dtype == 1

    def is_char(self):
        return self.dtype == 0

    def is_time(self):
        return self.dtype == 3

    def is_bool(self):
        return self.dtype == 4

    def is_set(self):
        return self.isSet == 1

    @classmethod
    def character(cls, size, length):
        base = (c_char * ((cls.CTRLBLOCK + size) * length))()
        data = (c_char * (size * length)).from_buffer(
            base, cls.CTRLBLOCK * BITSIZE["char"] * length
        )
        instance = cls(
            cls.DATATYPES_ENUM["char"],
            length,
            size,
            0,
            1,
            cast(base, c_void_p),
            cast(data, c_void_p),
        )
        return instance

    @classmethod
    def integer(cls, size):
        base = (c_int * (cls.CTRLBLOCK + size))()
        data = (c_int * size).from_buffer(base, cls.CTRLBLOCK * BITSIZE["int"])
        instance = cls(
            cls.DATATYPES_ENUM["int"],
            0,
            size,
            0,
            1,
            cast(base, c_void_p),
            cast(data, c_void_p),
        )
        return instance

    @classmethod
    def double(cls, size):
        base = (c_double * (cls.CTRLBLOCK + size))()
        data = (c_double * size).from_buffer(base, cls.CTRLBLOCK * BITSIZE["double"])
        instance = cls(
            cls.DATATYPES_ENUM["double"],
            0,
            size,
            0,
            1,
            cast(base, c_void_p),
            cast(data, c_void_p),
        )
        return instance

    @classmethod
    def bool(cls, size):
        base = (c_int * (cls.CTRLBLOCK + size))()
        data = (c_int * size).from_buffer(base, cls.CTRLBLOCK * BITSIZE["bool"])
        instance = cls(
            cls.DATATYPES_ENUM["bool"],
            0,
            size,
            0,
            1,
            cast(base, c_void_p),
            cast(data, c_void_p),
        )
        return instance

    @classmethod
    def time(cls, size):
        base = (c_int * (cls.CTRLBLOCK + size))()
        data = (c_int * size).from_buffer(base, cls.CTRLBLOCK * BITSIZE["time"])
        instance = cls(
            cls.DATATYPES_ENUM["time"],
            0,
            size,
            0,
            1,
            cast(base, c_void_p),
            cast(data, c_void_p),
        )
        return instance

    def __len__(self):
        return self.card

    def __iter__(self):
        getter = SpiceCell.DATATYPES_GET[self.dtype]
        length, card, data = self.length, self.card, self.data
        for i in range(card):
            yield getter(data, i, length)

    def __contains__(self, key):
        return key in self.__iter__()

    def __getitem__(self, key):
        getter = SpiceCell.DATATYPES_GET[self.dtype]
        if isinstance(key, slice):
            # TODO Typechecking
            if self.card == 0:
                return []
            else:
                start, stop, step = key.indices(self.card)
                return [
                    getter(self.data, i, self.length) for i in range(start, stop, step)
                ]
        elif key in range(-self.card, self.card):
            index = key if key >= 0 else self.card - abs(key)
            return getter(self.data, index, self.length)
        elif not isinstance(key, int):
            raise TypeError(
                "SpiceCell indices must be integers, not {}".format(type(key))
            )
        else:
            raise IndexError("SpiceCell index out of range")

    def reset(self):
        self.card = 0
        self.init = 0

    def __eq__(self, other):
        """
        element wise equality, other can be a list or cell
        I think sets should not equal a non set even if
        elements are equal... might be a bad idea
        :param other:
        :return:
        """
        if not hasattr(other, "__iter__"):
            return False
        if len(self) != len(other):
            return False
        if isinstance(other, SpiceCell):
            if other.dtype != self.dtype:
                return False
            if other.isSet != self.isSet:
                return False
        for x, y in zip(self, other):
            if x != y:
                return False
        return True


SpiceCellPointer = POINTER(SpiceCell)


def SPICEDOUBLE_CELL(size: int) -> SpiceCell:
    """
    Returns a Double Spice Cell with a given size
    :param size: number of elements
    :return: empty Spice Cell
    """
    return SpiceCell.double(size)


def SPICEINT_CELL(size: int) -> SpiceCell:
    """
    Returns a Int Spice Cell with a given size
    :param size: number of elements
    :return: empty Spice Cell
    """
    return SpiceCell.integer(size)


def SPICECHAR_CELL(size: int, length: int) -> SpiceCell:
    """
    Returns a Char Spice Cell with a given size
    :param size: number of elements
    :param length: width of elements
    :return: empty Spice Cell
    """
    return SpiceCell.character(size, length)


def SPICEBOOL_CELL(size: int) -> SpiceCell:
    """
    Returns a Bool Spice Cell with a given size
    :param size: number of elements
    :return: empty Spice Cell
    """
    return SpiceCell.bool(size)


def SPICETIME_CELL(size: int):
    """
    Returns a Time Spice Cell with a given size
    :param size: number of elements
    :return: empty Spice Cell
    """
    return SpiceCell.time(size)


# Spice Cell classes


class Cell_Time(SpiceCell):
    def __init__(self, size: int) -> None:
        """
        Init a Time Spice Cell with a given size and length
        :param size: number of elements
        """
        base = (c_int * (6 + size))()
        data = (c_int * size).from_buffer(base, 6 * BITSIZE["time"])
        super(Cell_Time, self).__init__(
            3, 0, size, 0, 1, cast(base, c_void_p), cast(data, c_void_p)
        )


class Cell_Bool(SpiceCell):
    def __init__(self, size: int) -> None:
        """
        Init a Bool Spice Cell with a given size and length
        :param size: number of elements
        """
        base = (c_int * (6 + size))()
        data = (c_int * size).from_buffer(base, 6 * BITSIZE["bool"])
        super(Cell_Bool, self).__init__(
            4, 0, size, 0, 1, cast(base, c_void_p), cast(data, c_void_p)
        )


class Cell_Int(SpiceCell):
    def __init__(self, size: int) -> None:
        """
        Init a Int Spice Cell with a given size and length
        :param size: number of elements
        """
        base = (c_int * (6 + size))()
        data = (c_int * size).from_buffer(base, 6 * BITSIZE["int"])
        super(Cell_Int, self).__init__(
            2, 0, size, 0, 1, cast(base, c_void_p), cast(data, c_void_p)
        )


class Cell_Double(SpiceCell):
    def __init__(self, size: int) -> None:
        """
        Init a Double Spice Cell with a given size and length
        :param size: number of elements
        """
        base = (c_double * (6 + size))()
        data = (c_double * size).from_buffer(base, 6 * BITSIZE["double"])
        super(Cell_Double, self).__init__(
            1, 0, size, 0, 1, cast(base, c_void_p), cast(data, c_void_p)
        )


class Cell_Char(SpiceCell):
    def __init__(self, size: int, length: int) -> None:
        """
        Init a Char Spice Cell with a given size and length
        :param size: number of elements
        :param length: width of elements
        """
        base = (c_char * ((6 + size) * length))()
        data = (c_char * (size * length)).from_buffer(
            base, 6 * BITSIZE["char"] * length
        )
        super(Cell_Char, self).__init__(
            0, length, size, 0, 1, cast(base, c_void_p), cast(data, c_void_p)
        )



"""
The MIT License (MIT)

Copyright (c) [2015-2022] [Andrew Annex]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import functools
from ctypes import c_int, c_double, c_char_p, POINTER, CFUNCTYPE, byref
from .support_types import SpiceCell, SpiceCellPointer, to_python_string
from typing import Callable, Union

UDFUNC = CFUNCTYPE(None, c_double, POINTER(c_double))
UDFUNS = CFUNCTYPE(None, c_double, POINTER(c_double))
UDFUNB = CFUNCTYPE(None, UDFUNS, c_double, POINTER(c_int))
UDSTEP = CFUNCTYPE(None, c_double, POINTER(c_double))
UDREFN = CFUNCTYPE(None, c_double, c_double, c_int, c_int, POINTER(c_double))
UDREPI = CFUNCTYPE(None, POINTER(SpiceCell), c_char_p, c_char_p)
UDREPU = CFUNCTYPE(None, c_double, c_double, c_double)
UDREPF = CFUNCTYPE(None)
UDBAIL = CFUNCTYPE(c_int)


def SpiceUDFUNC(f: Callable[[float], float]) -> UDFUNC:
    """
    Decorator for wrapping python functions in spice udfunc callback type

    :param f: function that has one argument of type float, and returns a float
    :return: wrapped udfunc function
    """

    @functools.wraps(f)
    def wrapping_udfunc(x: float, value: POINTER(c_double)) -> None:
        result = f(x)
        value[0] = c_double(result)

    return UDFUNC(wrapping_udfunc)


def SpiceUDFUNS(f: Callable[[float], float]) -> UDFUNS:
    """
    Decorator for wrapping python functions in spice udfuns callback type

    :param f: function that has one argument of type float, and returns a float
    :return: wrapped udfunc function
    """

    @functools.wraps(f)
    def wrapping_udfuns(x: float, value: POINTER(c_double)) -> None:
        result = f(x)
        value[0] = c_double(result)

    return UDFUNS(wrapping_udfuns)


def SpiceUDFUNB(f: Callable[[UDFUNS, float], int]) -> UDFUNB:
    """
    Decorator for wrapping python functions in spice udfunb callback type

    :param f: function to be wrapped
    :return: wrapped udfunb function
    """

    @functools.wraps(f)
    def wrapping_udfunb(udf: UDFUNS, et: float, xbool: POINTER(c_int)) -> None:
        # casting to bool fixes https://github.com/numpy/numpy/issues/14397
        result = bool(f(udf, et))
        xbool[0] = c_int(result)

    return UDFUNB(wrapping_udfunb)


def SpiceUDSTEP(f: Callable[[float], float]) -> UDSTEP:
    """
    Decorator for wrapping python functions in spice udstep callback type

    :param f: function to be wrapped
    :return: wrapped udstep function
    """

    @functools.wraps(f)
    def wrapping_udstep(x: float, value: POINTER(c_double)) -> None:
        result = f(x)
        value[0] = c_double(result)

    return UDSTEP(wrapping_udstep)


def SpiceUDREFN(
    f: Callable[[float, float, Union[bool, int], Union[bool, int]], float]
) -> UDREFN:
    """
    Decorator for wrapping python functions in spice udrefn callback type

    :param f: function to be wrapped
    :return: wrapped udrefn function
    """

    @functools.wraps(f)
    def wrapping_udrefn(
        t1: float,
        t2: float,
        s1: Union[bool, int],
        s2: Union[bool, int],
        t: POINTER(c_double),
    ) -> None:
        result = f(t1, t2, s1, s2)
        t[0] = c_double(result)

    return UDREFN(wrapping_udrefn)


def SpiceUDREPI(
    f: Callable[[Union[SpiceCell, SpiceCellPointer], str, str], None]
) -> UDREPI:
    """
    Decorator for wrapping python functions in spice udfrepi callback type

    :param f: function to be wrapped
    :return: wrapped udrepi function
    """

    @functools.wraps(f)
    def wrapping_udrepi(
        cnfine: Union[SpiceCell, SpiceCellPointer], srcpre: bytes, srcsurf: bytes
    ) -> None:
        f(cnfine, to_python_string(srcpre), to_python_string(srcsurf))

    return UDREPI(wrapping_udrepi)


def SpiceUDREPU(f: Callable[[float, float, float], None]) -> UDREPU:
    """
    Decorator for wrapping python functions in spice udrepu callback type

    :param f: function to be wrapped
    :return: wrapped udrepu function
    """

    @functools.wraps(f)
    def wrapping_udrepu(beg: float, end: float, et: float) -> None:
        f(beg, end, et)

    return UDREPU(wrapping_udrepu)


def SpiceUDREPF(f: Callable) -> UDREPF:
    """
    Decorator for wrapping python functions in spice udrepf callback type

    :param f: function to be wrapped
    :return: wrapped udrepf function
    """

    @functools.wraps(f)
    def wrapping_udrepf() -> None:
        f()

    return UDREPF(wrapping_udrepf)


def SpiceUDBAIL(f: Callable[[], Union[bool, int]]) -> UDBAIL:
    """
    Decorator for wrapping python functions in spice udbail callback type

    :param f: function to be wrapped
    :return: wrapped udbail function
    """

    @functools.wraps(f)
    def wrapping_udbail() -> int:
        result = f()
        return int(result)

    return UDBAIL(wrapping_udbail)


def CallUDFUNS(f: UDFUNS, x: float) -> float:
    """
    We are given a UDF CFUNCTYPE and want to call it in python

    :param f: SpiceUDFUNS
    :param x: some scalar
    :return: value
    """
    value = c_double()
    f(x, byref(value))
    return value.value
